*************************************************************************************
*														*
*	Maze generator for EASy68K V1.00	12/8/2008						*
*														*
*	Generates random mazes and displays them in the window.				*
*														*
*	As the path that the generator takes is stored on the stack quite a bit		*
*	of stack space is needed e.g. in the case of an 80 x 60 cell maze about		*
*	4K of stack space is required. The maximum that would ever be needed is		*
*	a little under x_sixe * y_size * 8 bytes which for an 80 x 60 cell maze		*
*	is almost 38K but it's very unlikely that this maximum amount will ever		*
*	be needed.												*
*														*
*	Once a maze is displayed hitting any key will clear the screen and a new	*
*	maze will be generated.										*
*														*
*														*
*	More 68000 and other projects can be found on my website at ..			*
*														*
*	 http://members.lycos.co.uk/leeedavison/index.html					*
*														*
*	mail : leeedavison@lycos.co.uk								*
*														*
*************************************************************************************
*
* first some constants

x_size	EQU 80				* maze width
y_size	EQU 60				* maze height


*************************************************************************************
*
* the code

	ORG		$1000

start
	BSR		Initialise			* go setup everything

* main loop

loop
	MOVE.w	#$FF00,d1			* clear screen
	MOVEQ		#11,d0			* position cursor
	TRAP		#15

	BSR		clear_maze			* clear the maze array
	BSR		make_maze			* make the maze
	BSR		draw_maze			* draw the maze
wait_key
	MOVEQ		#7,d0				* read key status
	TRAP		#15

	TST.b		d1				* test the result
	BEQ.s		wait_key			* if no key go wait some more

	MOVEQ		#5,d0				* read a key
	TRAP		#15

	BRA.s		loop				* run again


*************************************************************************************
*
* draw the maze. a not-at-all optimised draw routine. for each cell each wall bit is
* tested in turn and, if the bit is set, the wall is drawn.

draw_maze
	MOVEQ		#84,d0			* draw a line from x1,y1 to x2,y2
	MOVEQ		#0,d5				* clear the maze index
	MOVEQ		#0,d7				* clear the y co-ordinate
draw_y_loop
	MOVEQ		#0,d6				* clear the x co-ordinate
draw_x_loop
	BTST		#2,(a6,d5.l)		* test top wall bit
	BEQ.s		no_top_wall			* if no top wall skip top wall draw

							* draw x1,y1 to x2,y1
	MOVE.w	d6,d1				* x1 = x
	MOVE.w	d7,d2				* y1 = y
	MOVE.w	d6,d3				* x2 = x
	ADDQ.w	#7,d3				* x2 = x + 1
	MOVE.w	d7,d4				* y2 = y
	TRAP		#15

no_top_wall
	BTST		#1,(a6,d5.l)		* test right wall bit
	BEQ.s		no_right_wall		* if no right wall skip right wall draw

							* draw x2,y1 to x2,y2
	MOVE.w	d6,d1				* x1 = x
	ADDQ.w	#7,d1				* x2 = x1 + 1
	MOVE.w	d7,d2				* y1 = y
	MOVE.w	d6,d3				* x2 = x
	ADDQ.w	#7,d3				* x2 = x1 + 1
	MOVE.w	d7,d4				* y2 = y
	ADDQ.w	#7,d4				* y2 = y1 + 1
	TRAP		#15

no_right_wall
	BTST		#0,(a6,d5.l)		* test bottom wall bit
	BEQ.s		no_bottom_wall		* if no bottom wall skip bottom wall draw

							* draw x2,y2 to x1,y2
	MOVE.w	d6,d1				* x1 = x
	ADDQ.w	#7,d1				* x2 = x1 + 1
	MOVE.w	d7,d2				* y1 = y
	ADDQ.w	#7,d2				* y2 = y1 + 1
	MOVE.w	d6,d3				* x2 = x
	MOVE.w	d7,d4				* y2 = y
	ADDQ.w	#7,d4				* y2 = y1 + 1
	TRAP		#15

no_bottom_wall
	BTST		#3,(a6,d5.l)		* test left wall bit
	BEQ.s		no_left_wall		* if no left wall skip left wall draw

							* draw x1,y2 to x1,y1
	MOVE.w	d6,d1				* x1 = x
	MOVE.w	d7,d2				* y1 = y
	ADDQ.w	#7,d2				* y2 = y1 + 1
	MOVE.w	d6,d3				* x2 = x
	MOVE.w	d7,d4				* y2 = y
	TRAP		#15

no_left_wall
	ADDQ.l	#1,d5				* increment the array index

	ADDQ.w	#8,d6				* increment the x co-ordinate
	CMPI.w	#x_size*8,d6		* compare with max
	BNE		draw_x_loop			* loop if more to do

	ADDQ.w	#8,d7				* increment the y co-ordinate
	CMPI.w	#y_size*8,d7		* compare with max
	BNE		draw_y_loop			* loop if more to do

	RTS


*************************************************************************************
*
* make the maze. this is done by chosing a random start x,y and pushing this on the
* stack as the parameters for chose_cell. chose_cell will then not return until the
* maze is complete.

make_maze
	MOVE.w	#y_size,d5			* get the range for the y start
	BSR		rand_d5			* get a random value in the range 0 to
							* (d2.w - 1)
	MOVE.w	d0,-(sp)			* push start y

	MOVE.w	#x_size,d5			* get the range for the x start
	BSR		rand_d5			* get a random value in the range 0 to
							* (d2.w - 1)
	MOVE.w	d0,-(sp)			* push start x
	BSR		chose_cell			* go build the maze
	ADDQ.w	#4,sp				* dump x,y
	RTS


*************************************************************************************
*
* test a cell to see if it's free. the free cell bitfield is shifted left and b0 will
* be set if x,y is inside the array and the cell at x,y is free

test_cell
	ADD.w		d3,d3				* shift the free cell bitfield

	TST.w		d2				* test this cell y
	BMI.s		exit_test_cell		* if < min skip flag free

	CMP.w		#y_size,d2			* compare with y size
	BGE.s		exit_test_cell		* if > max skip flag free

	TST.w		d1				* test this cell x
	BMI.s		exit_test_cell		* if < min skip flag free

	CMP.w		#x_size,d1			* compare with x size
	BGE.s		exit_test_cell		* if > max skip flag free

							* now calculate offset to this cell's byte
	EXT.l		d1				* clear this cell y high word
	MOVE.w	d2,d0				* copy this cell x
	MULU.w	#x_size,d0			* y * x size
	ADD.l		d1,d0				* y * x size + x
	CMPI.b	#$0F,(a6,d0.l)		* compare the cell with unvisited value
	BNE.s		exit_test_cell		* if used skip flag free

	ADDQ.w	#1,d3				* set b0 to flag this cell is free
exit_test_cell
	RTS


*************************************************************************************
*
* chose the next cell. if there is an adjacent free cell the co-ordinates are pushed
* and then the routine calls itself. eventually a cell is reached that is a dead end
* and at this point the routine returns. each time this happens it effectively steps
* backwards along the path previously taken until a cell with at least one adjacent
* free cell is found or the original cell is reached at which time chose_cell finally
* exits.

chose_cell_x	EQU 4				* cell x co-ordiante
chose_cell_y	EQU 6				* cell y co-ordiante

chose_cell
	MOVEM.w	chose_cell_x(sp),d1-d2	* get this cell x,y. clears the high word
chose_cell_2
	MOVEQ		#0,d3				* flag no cells free

	SUBQ.w	#1,d1				* x - 1,y
	BSR		test_cell			* test if cell is empty

	ADDQ.w	#1,d1				* x
	SUBQ.w	#1,d2				* y - 1
	BSR		test_cell			* test if cell is empty

	ADDQ.w	#1,d1				* x + 1
	ADDQ.w	#1,d2				* y
	BSR		test_cell			* test if cell is empty

	SUBQ.w	#1,d1				* x
	ADDQ.w	#1,d2				* y + 1
	BSR		test_cell			* test if cell is empty

	TST.w		d3				* test for adjacent free cells
	BEQ.s		exit_chose_cell		* if no free cells just exit

	SUBQ.w	#1,d2				* y

* now chose a cell from those available

	LEA		free_01(pc),a1		* point to the free cell table
	SUBQ.w	#1,d3				* make 1 to n into 0 to n - 1
	MOVE.w	d3,d4				* copy the free cell index
	ASL.w		#2,d4				* * 4
	ADD.w		d4,d3				* * 5
	ADD.w		d3,d3				* * 10 bytes per free cell
	ADD.w		d3,a1				* point to this cell's free table
	MOVE.w	(a1)+,d5			* get the number of directions available
	BMI.s		no_choice			* if -ve use the only available direction

	BSR		rand_d5			* get a random value in the range 0 to
							* (d5.w - 1)
	ADD.w		d0,d0				* * 2 to index the word table
	ADD.w		d0,a1				* get the index for the chosen offset
no_choice
	MOVE.w	(a1),d0			* get the offset for the chosen direction

							* calculate array index for this cell's byte
	MOVE.w	d2,d4				* copy this cell y
	MULU.w	#x_size,d4			* y * x size
	ADD.l		d1,d4				* y * x size + x

	LEA		next_offset(pc),a1	* point to the direction tables
	MOVE.w	this_mask(a1,d0.w),d3	* get this cell mask
	AND.b		d3,(a6,d4.l)		* mask this cell

	MOVE.w	(a1,d0.w),d3		* get the offset to the next cell
	EXT.l		d3				* make into a longword offset
	ADD.l		d3,d4				* add this to the current cell index

	MOVE.w	next_mask(a1,d0.w),d3	* get the next cell mask
	AND.b		d3,(a6,d4.l)		* mask the next cell

	ADD.w		next_dx(a1,d0.w),d1	* add the delta x to the next cell to x
	ADD.w		next_dy(a1,d0.w),d2	* add the delta y to the next cell to y
	TST.w		d5				* test if this was a forced move
	BMI.s		forced_move			* if this is a forced move don't bother to
							* save the return point

	MOVEM.w	d1-d2,-(sp)			* push x,y. this is stepping forwards along
							* the path as it's built
	BSR		chose_cell			* continue building the maze
	ADDQ.w	#4,sp				* dump x,y. this is stepping back along the
							* path looking for adjacent free cells

	BRA		chose_cell			* continue building the maze

* if, during the forward path, a cell is reached that has only one free direction
* then that cell's x,y is not preserved as there is no point stepping back through
* it looking for another possible path

forced_move
	MOVEM.w	d1-d2,chose_cell_x(sp)	* just update x,y. overwrite the previous path
	BRA		chose_cell_2		* continue building the maze

* reaching this point means that no surrounding cells are free so step back along the
* path by returning to the calling point and discarding an x,y pair from the stack

exit_chose_cell
	RTS


*************************************************************************************
*
* offsets and masks for moving through the maze removing walls as progress is made

next_offset
	dc.w	x_size				* offset in array to cell below
	dc.w	1					* offset in array to cell right
	dc.w	-x_size				* offset in array to cell above
	dc.w	-1					* offset in array to cell left

this_mask					EQU	*-next_offset
	dc.w	$FFFE					* mask for bit 0
	dc.w	$FFFD					* mask for bit 1
next_mask					EQU	*-next_offset
	dc.w	$FFFB					* mask for bit 2
	dc.w	$FFF7					* mask for bit 3
	dc.w	$FFFE					* mask for bit 0
	dc.w	$FFFD					* mask for bit 1

* deltas for x,y to the next cell

next_dx					EQU	*-next_offset
	dc.w	0					*
next_dy					EQU	*-next_offset
	dc.w	1					*
	dc.w	0					*
	dc.w	-1					*
	dc.w	0					*


*************************************************************************************
*
* each cell is arranged as follows. a set bit indicates that a wall is present so an
* unvisited cell will have a value of $xF
*
*		x  ->
*
*	y	+-----------------+-----------------+-----------------+
*		|			| 	bit 2		|			|
*	|	|			| b		    b	|			|
*	V	|			| i		    i	|			|
*		|			| t	x,y-1	    t	|			|
*		|			|			|			|
*		|			| 3		    1	|			|
*		|			|	bit 0		|			|
*		+-----------------+-----------------+-----------------+
*		|	bit 2		|	bit 2		|	bit 2		|
*		| b		    b	| b		    b	| b		    b	|
*		| i		    i	| i		    i	| i		    i	|
*		| t	x-1,y	    t	| t	 x,y	    t	| t	x+1,y	    t	|
*		|			|			|			|
*		| 3		    1	| 3		    1	| 3		    1	|
*		|	bit 0		|	bit 0		|	bit 0		|
*		+-----------------+-----------------+-----------------+
*		|			|	bit 2		|			|
*		|			| b		    b	|			|
*		|			| i		    i	|			|
*		|			| t	x,y+1	    t	|			|
*		|			|			|			|
*		|			| 3		    1	|			|
*		|			|	bit 0		|			|
*		+-----------------+-----------------+-----------------+
*
*
*************************************************************************************
*
* directions available. first word is the number of directions available. where only
* one direction is free a count of -1 flags a forced choice. the following words are
* the offset into the above tables for the masks, deltas and offset
*
* if, when chosing a new direction, the method of chosing any path and then stepping
* round until a free path is found tends to make mazes that spiral in whichever
* direction is taken when stepping. to avoid this deocil or widdershins bias this
* table is used so that the next cell is chosen from only the free cells available

free_01
	dc.w	-1					* adjacent cells free, -1 shows forced choice
	dc.w	0					* bit 0 direction free
	dc.w	-1					* filler
	dc.w	-1					* filler
	dc.w	-1					* filler
free_02
	dc.w	-1					* adjacent cells free, -1 shows forced choice
	dc.w	2					* bit 1 direction free
	dc.w	-1					* filler
	dc.w	-1					* filler
	dc.w	-1					* filler
free_03
	dc.w	2					* adjacent cells free
	dc.w	0					* bit 0 direction free
	dc.w	2					* bit 1 direction free
	dc.w	-1					* filler
	dc.w	-1					* filler
free_04
	dc.w	-1					* adjacent cells free, -1 shows forced choice
	dc.w	4					* bit 2 direction free
	dc.w	-1					* filler
	dc.w	-1					* filler
	dc.w	-1					* filler
free_05
	dc.w	2					* adjacent cells free
	dc.w	0					* bit 0 direction free
	dc.w	4					* bit 2 direction free
	dc.w	-1					* filler
	dc.w	-1					* filler
free_06
	dc.w	2					* adjacent cells free
	dc.w	2					* bit 1 direction free
	dc.w	4					* bit 2 direction free
	dc.w	-1					* filler
	dc.w	-1					* filler
free_07
	dc.w	3					* adjacent cells free
	dc.w	0					* bit 0 direction free
	dc.w	2					* bit 1 direction free
	dc.w	4					* bit 2 direction free
	dc.w	-1					* filler
free_08
	dc.w	-1					* adjacent cells free, -1 shows forced choice
	dc.w	6					* bit 3 direction free
	dc.w	-1					* filler
	dc.w	-1					* filler
	dc.w	-1					* filler
free_09
	dc.w	2					* adjacent cells free
	dc.w	0					* bit 0 direction free
	dc.w	6					* bit 3 direction free
	dc.w	-1					* filler
	dc.w	-1					* filler
free_0A
	dc.w	2					* adjacent cells free
	dc.w	2					* bit 1 direction free
	dc.w	6					* bit 3 direction free
	dc.w	-1					* filler
	dc.w	-1					* filler
free_0B
	dc.w	3					* adjacent cells free
	dc.w	0					* bit 0 direction free
	dc.w	2					* bit 1 direction free
	dc.w	6					* bit 3 direction free
	dc.w	-1					* filler
free_0C
	dc.w	2					* adjacent cells free
	dc.w	4					* bit 2 direction free
	dc.w	6					* bit 3 direction free
	dc.w	-1					* filler
	dc.w	-1					* filler
free_0D
	dc.w	3					* adjacent cells free
	dc.w	0					* bit 0 direction free
	dc.w	4					* bit 2 direction free
	dc.w	6					* bit 3 direction free
	dc.w	-1					* filler
free_0E
	dc.w	3					* adjacent cells free
	dc.w	2					* bit 1 direction free
	dc.w	4					* bit 2 direction free
	dc.w	6					* bit 3 direction free
	dc.w	-1					* filler
free_0F
	dc.w	4					* adjacent cells free
	dc.w	0					* bit 0 direction free
	dc.w	2					* bit 1 direction free
	dc.w	4					* bit 2 direction free
	dc.w	6					* bit 3 direction free


*************************************************************************************
*
* setup stuff

Initialise
	LEA		variables(pc),a3		* get the variables base address

	MOVEQ		#8,d0				* get time in 1/100 ths seconds
	TRAP		#15

	EORI.l	#$DEADBEEF,d1		* EOR with the initial PRNG seed, this must
							* result in any value but zero
	MOVE.l	d1,PRNlword(a3)		* save the initial PRNG seed

	LEA		array(a3),a6		* a6 points to the array
	RTS


*************************************************************************************
*
* clear the maze array. an unvisited cell has all four walls present and so will have
* a value of $xF

clear_maze
	MOVE.w	#x_size,d7			* get the x_size
	MULU.w	#y_size,d7			* * the y_size
	SUBQ.l	#4,d7				* adjust for loop type

	MOVE.l	#$0F0F0F0F,d0		* set the longword
fill_loop
	MOVE.l	d0,(a6,d7.l)		* fill the longword
	SUBQ.l	#4,d7				* decrement byte count in longwords
	BPL.s		fill_loop			* loop if more to do

	RTS


*************************************************************************************
*
* generate a random value in d0.w in range 0 to (d5.w - 1)

rand_d5
	BSR.s		random_n			* generate a random number
	MOVEQ		#0,d0				* clear the longword
	MOVE.w	PRNlword(a3),d0		* get a random word. doing this instead of
							* just getting a longword ensures that the
							* following DIVU never overflows
	DIVU.w	d5,d0				* divide by range
	SWAP		d0				* use the signed remainder
	RTS


*************************************************************************************
*
* This is the code that generates the pseudo random sequence. A seed word located in
* PRNlword(a3) is loaded into a register before being operated on to generate the
* next number in the sequence. This number is then saved as the seed for the next
* time it's called.
*
* This code is adapted from the 32 bit version of RND(n) used in EhBASIC68. Taking
* the 19th next number is slower but helps to hide the shift and add nature of this
* generator as can be seen from analysing the output.

random_n
	MOVEM.l	d0-d2,-(sp)			* save d0, d1 and d2
	MOVE.l	PRNlword(a3),d0		* get current seed longword
	MOVEQ		#$AF-$100,d1		* set EOR value
	MOVEQ		#18,d2			* do this 19 times
Ninc0
	ADD.l		d0,d0				* shift left 1 bit
	BCC.s		Ninc1				* if bit not set skip feedback

	EOR.b		d1,d0				* do Galois LFSR feedback
Ninc1
	DBF		d2,Ninc0			* loop

	MOVE.l	d0,PRNlword(a3)		* save back to seed longword
	MOVEM.l	(sp)+,d0-d2			* restore d0, d1 and d2

	RTS


*************************************************************************************
*
* variables

variables

	OFFSET	0				* going to use relative addressing

PRNlword
	ds.l	1					* PRNG seed long word

array
	ds.b	x_size*y_size			* make space for the array
	ds.w	0					* ensure even


*************************************************************************************

	END		start

*************************************************************************************
